"""J.A.R.V.I.S. Utility Cog."""
import re
import secrets
import string
from io import BytesIO

import numpy as np
from dis_snek import InteractionContext, Scale, Snake, const
from dis_snek.models.discord.embed import Color, EmbedField
from dis_snek.models.discord.file import File
from dis_snek.models.discord.guild import Guild
from dis_snek.models.discord.role import Role
from dis_snek.models.discord.user import User
from dis_snek.models.snek.application_commands import (
    OptionTypes,
    SlashCommandChoice,
    slash_command,
    slash_option,
)
from dis_snek.models.snek.command import cooldown
from dis_snek.models.snek.cooldowns import Buckets
from PIL import Image

import jarvis
from jarvis.config import get_config
from jarvis.data import pigpen
from jarvis.data.robotcamo import emotes, hk, names
from jarvis.utils import build_embed, get_repo_hash

JARVIS_LOGO = Image.open("jarvis_small.png").convert("RGBA")


class UtilCog(Scale):
    """
    Utility functions for J.A.R.V.I.S.

    Mostly system utility functions, but may change over time
    """

    def __init__(self, bot: Snake):
        self.bot = bot
        self.config = get_config()

    @slash_command(name="Status", description="Retrieve J.A.R.V.I.S. status")
    @cooldown(bucket=Buckets.CHANNEL, rate=1, interval=30)
    async def _status(self, ctx: InteractionContext) -> None:
        title = "J.A.R.V.I.S. Status"
        desc = "All systems online"
        color = "#3498db"
        fields = []

        fields.append(EmbedField(name="discord.py", value=const.__version__))
        fields.append(EmbedField(name="Version", value=jarvis.__version__, inline=False))
        fields.append(EmbedField(name="Git Hash", value=get_repo_hash()[:7], inline=False))
        embed = build_embed(title=title, description=desc, fields=fields, color=color)
        await ctx.send(embed=embed)

    @slash_command(
        name="logo",
        description="Get the current logo",
    )
    @cooldown(bucket=Buckets.CHANNEL, rate=1, interval=30)
    async def _logo(self, ctx: InteractionContext) -> None:
        with BytesIO() as image_bytes:
            JARVIS_LOGO.save(image_bytes, "PNG")
            image_bytes.seek(0)
            logo = File(image_bytes, filename="logo.png")
            await ctx.send(file=logo)

    @slash_command(name="rchk", description="Robot Camo HK416")
    async def _rchk(self, ctx: InteractionContext) -> None:
        await ctx.send(content=hk)

    @slash_command(
        name="rcauto",
        description="Automates robot camo letters",
    )
    @slash_option(
        name="text",
        description="Text to camo-ify",
        option_type=OptionTypes.STRING,
        required=True,
    )
    async def _rcauto(self, ctx: InteractionContext, text: str) -> None:
        to_send = ""
        if len(text) == 1 and not re.match(r"^[A-Z0-9-()$@!?^'#. ]$", text.upper()):
            await ctx.send("Please use ASCII characters.", hidden=True)
            return
        for letter in text.upper():
            if letter == " ":
                to_send += "      "
            elif re.match(r"^[A-Z0-9-()$@!?^'#.]$", letter):
                id = emotes[letter]
                if ctx.author.is_on_mobile():
                    to_send += f":{names[id]}:"
                else:
                    to_send += f"<:{names[id]}:{id}>"
        if len(to_send) > 2000:
            await ctx.send("Too long.", hidden=True)
        else:
            await ctx.send(to_send)

    @slash_command(name="avatar", description="Get a user avatar")
    @slash_option(
        name="user",
        description="User to view avatar of",
        option_type=OptionTypes.USER,
        required=False,
    )
    @cooldown(bucket=Buckets.USER, rate=1, interval=5)
    async def _avatar(self, ctx: InteractionContext, user: User = None) -> None:
        if not user:
            user = ctx.author

        avatar = user.display_avatar
        embed = build_embed(title="Avatar", description="", fields=[], color="#00FFEE")
        embed.set_image(url=avatar)
        embed.set_author(name=f"{user.username}#{user.discriminator}", icon_url=avatar)
        await ctx.send(embed=embed)

    @slash_command(
        name="roleinfo",
        description="Get role info",
    )
    @slash_option(
        name="role",
        description="Role to get info of",
        option_type=OptionTypes.ROLE,
        required=True,
    )
    async def _roleinfo(self, ctx: InteractionContext, role: Role) -> None:
        fields = [
            EmbedField(name="ID", value=role.id),
            EmbedField(name="Name", value=role.name),
            EmbedField(name="Color", value=str(role.color)),
            EmbedField(name="Mention", value=f"`{role.mention}`"),
            EmbedField(name="Hoisted", value="Yes" if role.hoist else "No"),
            EmbedField(name="Position", value=str(role.position)),
            EmbedField(name="Mentionable", value="Yes" if role.mentionable else "No"),
            EmbedField(name="Member Count", value=role.members),
        ]
        embed = build_embed(
            title="",
            description="",
            fields=fields,
            color=Color.from_hex(role.color),
            timestamp=role.created_at,
        )
        embed.set_footer(text="Role Created")
        embed.set_thumbnail(url="attachment://color_show.png")

        data = np.array(JARVIS_LOGO)
        r, g, b, a = data.T

        fill = a > 0

        data[..., :-1][fill.T] = list(role.color.to_rgb())

        im = Image.fromarray(data)

        with BytesIO() as image_bytes:
            im.save(image_bytes, "PNG")
            image_bytes.seek(0)
            color_show = File(image_bytes, filename="color_show.png")

            await ctx.send(embed=embed, file=color_show)

    @slash_command(
        name="userinfo",
        description="Get user info",
    )
    @slash_option(
        name="user",
        description="User to get info of",
        option_type=OptionTypes.USER,
        required=False,
    )
    async def _userinfo(self, ctx: InteractionContext, user: User = None) -> None:
        if not user:
            user = ctx.author
        user_roles = user.roles
        if user_roles:
            user_roles = sorted(user.roles, key=lambda x: -x.position)
            _ = user_roles.pop(-1)
        fields = [
            EmbedField(
                name="Joined",
                value=user.joined_at.strftime("%a, %b %-d, %Y %-I:%M %p"),
            ),
            EmbedField(
                name="Registered",
                value=user.created_at.strftime("%a, %b %-d, %Y %-I:%M %p"),
            ),
            EmbedField(
                name=f"Roles [{len(user_roles)}]",
                value=" ".join([x.mention for x in user_roles]) if user_roles else "None",
                inline=False,
            ),
        ]

        embed = build_embed(
            title="",
            description=user.mention,
            fields=fields,
            color=Color.from_hex(str(user_roles[0].color) if user_roles else "#3498db"),
        )

        embed.set_author(
            name=f"{user.display_name}#{user.discriminator}", icon_url=user.display_avatar
        )
        embed.set_thumbnail(url=user.display_avatar)
        embed.set_footer(text=f"ID: {user.id}")

        await ctx.send(embed=embed)

    @slash_command(name="serverinfo", description="Get server info")
    async def _server_info(self, ctx: InteractionContext) -> None:
        guild: Guild = ctx.guild

        owner = (
            f"{guild.owner.display_name}#{guild.owner.discriminator}"
            if guild.owner
            else "||`[redacted]`||"
        )

        categories = len(guild.categories)
        text_channels = len(guild.text_channels)
        voice_channels = len(guild.voice_channels)
        members = guild.member_count
        roles = len(guild.roles)
        role_list = ", ".join(role.name for role in guild.roles)

        fields = [
            EmbedField(name="Owner", value=owner),
            EmbedField(name="Channel Categories", value=categories),
            EmbedField(name="Text Channels", value=text_channels),
            EmbedField(name="Voice Channels", value=voice_channels),
            EmbedField(name="Members", value=members),
            EmbedField(name="Roles", value=roles),
        ]
        if len(role_list) < 1024:
            fields.append(EmbedField(name="Role List", value=role_list, inline=False))

        embed = build_embed(title="", description="", fields=fields, timestamp=guild.created_at)

        embed.set_author(name=guild.name, icon_url=guild.icon_url)
        embed.set_thumbnail(url=guild.icon_url)
        embed.set_footer(text=f"ID: {guild.id} | Server Created")

        await ctx.send(embed=embed)

    @slash_command(
        name="pw",
        sub_cmd_name="gen",
        description="Generate a secure password",
        scopes=[862402786116763668],
    )
    @slash_option(
        name="length",
        description="Password length (default 32)",
        option_type=OptionTypes.INTEGER,
        required=False,
    )
    @slash_option(
        name="chars",
        description="Characters to include (default last option)",
        option_type=OptionTypes.INTEGER,
        required=False,
        options=[
            SlashCommandChoice(name="A-Za-z", value=0),
            SlashCommandChoice(name="A-Fa-f0-9", value=1),
            SlashCommandChoice(name="A-Za-z0-9", value=2),
            SlashCommandChoice(name="A-Za-z0-9!@#$%^&*", value=3),
        ],
    )
    @cooldown(bucket=Buckets.USER, rate=1, interval=15)
    async def _pw_gen(self, ctx: InteractionContext, length: int = 32, chars: int = 3) -> None:
        if length > 256:
            await ctx.send("Please limit password to 256 characters", hidden=True)
            return
        choices = [
            string.ascii_letters,
            string.hexdigits,
            string.ascii_letters + string.digits,
            string.ascii_letters + string.digits + "!@#$%^&*",
        ]

        pw = "".join(secrets.choice(choices[chars]) for i in range(length))
        await ctx.send(
            f"Generated password:\n`{pw}`\n\n"
            '**WARNING: Once you press "Dismiss Message", '
            "*the password is lost forever***",
            hidden=True,
        )

    @slash_command(name="pigpen", description="Encode a string into pigpen")
    @slash_option(
        name="text", description="Text to encode", option_type=OptionTypes.STRING, required=True
    )
    async def _pigpen(self, ctx: InteractionContext, text: str) -> None:
        outp = "`"
        for c in text:
            c = c.lower()
            if c.lower() in pigpen.lookup:
                c = pigpen.lookup[c.lower()]
            elif c == " ":
                c = "   "
            elif c == "`":
                continue
            outp += c + " "
        outp += "`"
        await ctx.send(outp[:2000])


def setup(bot: Snake) -> None:
    """Add UtilCog to J.A.R.V.I.S."""
    bot.add_cog(UtilCog(bot))
